/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                        LibWin.h                         **/
/**                                                         **/
/** This file contains Windows-dependent definitions and    **/
/** declarations for the emulation library.                 **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-2007                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifndef LIBWIN_H
#define LIBWIN_H
#ifdef __cplusplus
extern "C" {
#endif

#include <Windows.h>

//#define BPP16    /* Pixels are 16bit values in 5:5:5 format  */
#define BPS16    /* Samples are 16bit signed values          */

#define HOMEPAGE      "http://fms.komkon.org/"

#define ASK_SAVEFILE  0x01
#define ASK_ASSOCIATE 0x02

#define SND_MINBLOCK  64
#define SND_HEADERS   64

#define MODE_640x480  -2
#define MODE_320x240  -1

/** WMENU_* **************************************************/
/** Default menu actions supported by LibWin.               **/
/*************************************************************/
#define WMENU_QUIT     900
#define WMENU_SIZEx1   901
#define WMENU_SIZEx2   902
#define WMENU_SIZEx3   903
#define WMENU_SIZEx4   904
#define WMENU_FULL     905
#define WMENU_SNDOFF   906 /* Group */
#define WMENU_SNDMIDI  907
#define WMENU_SND11kHz 908
#define WMENU_SND22kHz 909
#define WMENU_SND44kHz 910
#define WMENU_SND10ms  912 /* Group */
#define WMENU_SND50ms  913
#define WMENU_SND100ms 914
#define WMENU_SND250ms 915
#define WMENU_SND500ms 916
#define WMENU_SAVECPU  918
#define WMENU_SNDLOG   919
#define WMENU_SaveMID  920
#define WMENU_HELP     921
#define WMENU_WEB      922
#define WMENU_AUTHOR   923

/** PIXEL() **************************************************/
/** Windows uses the 15bpp 5:5:5 pixel format.              **/
/*************************************************************/
#ifdef BPP16
#define PIXEL(R,G,B) (pixel)(((31*(R)/255)<<10)|((31*(G)/255)<<5)|(31*(B)/255))
#else
#define PIXEL(R,G,B) (pixel)(((7*(R)/255)<<5)|((7*(G)/255)<<2)|(3*(B)/255))
#endif

/** Application Variables ************************************/
/** Please define initial application appearance via these  **/
/** variables.                                              **/
/************************************ TO BE WRITTEN BY USER **/
extern const char *AppTitle;
extern const char *AppClass;
extern int AppWidth;
extern int AppHeight;
extern int AppMode;
extern int AppX;                   /* These two are provided */
extern int AppY;

/** Application() ********************************************/
/** This is the main application function you implement. It **/
/** is called from WinMain() after initializing things.     **/
/************************************ TO BE WRITTEN BY USER **/
int Application(int argc,char *argv[]);

/** WinGetAppInstance() **************************************/
/** Get application instance.                               **/
/*************************************************************/
HINSTANCE WinGetAppInstance(void);

/** WinGetAppWindow() ****************************************/
/** Get main window handle.                                 **/
/*************************************************************/
HWND WinGetAppWindow(void);

/** WinSetCmdHandler() ***************************************/
/** Attach a handler that will be called on menu selections **/
/** and other commands.                                     **/
/*************************************************************/
void WinSetCmdHandler(void (*Handler)(unsigned int Cmd));

/** WinProcessMsgs() *****************************************/
/** Call this function periodically to process Windows      **/
/** messages. Returns number of messages processed.         **/
/*************************************************************/
int WinProcessMsgs(void);

/** WinSetWindow() *******************************************/
/** Set window size and position. Returns 0 if failed.      **/
/*************************************************************/
int WinSetWindow(int X,int Y,int Width,int Height);

/** WinAskFilename() *****************************************/
/** Shows an open-file dialog and makes user select a file. **/
/** Returns 0 on failure. ATTENTION: This function returns  **/
/** pointer to the internal buffer!                         **/
/*************************************************************/
const char *WinAskFilename(const char *Title,const char *Types,unsigned int Options);

/** WinGetInteger() ******************************************/
/** Get integer setting by name or return Default if not    **/
/** found.                                                  **/
/*************************************************************/
unsigned int WinGetInteger(const char *Section,const char *Name,unsigned int Default);

/** WinSetInteger() ******************************************/
/** Set integer setting to a given value.                   **/
/*************************************************************/
void WinSetInteger(const char *Section,const char *Name,unsigned int Value);

/** WinGetString() *******************************************/
/** Get string setting by name.                             **/
/*************************************************************/
char *WinGetString(const char *Section,const char *Name,char *Value,int MaxChars);

/** WinSetString() *******************************************/
/** Set string setting to a given value.                    **/
/*************************************************************/
void WinSetString(const char *Section,const char *Name,const char *Value);

/** WinGetJoysticks() ****************************************/
/** Get the state of *real* joypad buttons (1="pressed"),   **/
/** but *not* the keyboard keys. Refer to BTN_* #defines    **/
/** for the button mappings.                                **/
/*************************************************************/
unsigned int WinGetJoysticks(void);

/** WinSetJoysticks() ****************************************/
/** Shows a dialog to set joystick button mappings.          */
/*************************************************************/
void WinSetJoysticks(unsigned int Buttons);

/** WinAssociate() *******************************************/
/** Associates files with given extension to the application */
/*************************************************************/
int WinAssociate(const char *Ext,const char *Comment,int Icon);

/** InitAudio() **********************************************/
/** Initialize sound. Returns rate (Hz) on success, else 0. **/
/** Rate=0 to skip initialization (will be silent). The     **/
/** Delay value is given in millseconds and determines how  **/
/** much buffering will be done.                            **/
/*************************************************************/
unsigned int InitAudio(unsigned int Rate,unsigned int Delay);

/** TrashAudio() *********************************************/
/** Free resources allocated by InitAudio().                **/
/*************************************************************/
void TrashAudio(void);

#ifdef __cplusplus
}
#endif
#endif /* LIBWIN_H */
