/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                         SndMSDOS.c                      **/
/**                                                         **/
/** This file contains standard sound generation routines   **/
/** for MSDOS using SoundBlaster and Adlib.                 **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-2007                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#include "EMULib.h"
#include "Sound.h"

static struct
{
  int Type;                       /* Channel type (SND_*)             */
  int Freq;                       /* Channel frequency (Hz)           */
  int Volume;                     /* Channel volume (0..255)          */

  const signed char *Data;        /* Wave data (-128..127 each)       */
  int Length;                     /* Wave length in Data              */
  int Rate;                       /* Wave playback rate (or 0Hz)      */
  int Pos;                        /* Wave current position in Data    */  

  int Count;                      /* Phase counter                    */
} CH[SND_CHANNELS];

static int SndRate=0;             /* Sound rate (1=Adlib, 0=Off)      */
int MasterSwitch;                 /* Switches to turn channels on/off */
int MasterVolume;                 /* Master volume                    */

static int NoiseGen;              /* Noise generator seed             */
static int Wave[SND_BUFSIZE];     /* Buffer where wave mixing occurs  */
static sample OutBuf[SND_BUFSIZE];/* Output buffer for WriteAudio()   */

static const signed char *DOSGetWave(int Channel);
static void DOSSetWave(int Channel,const signed char *Data,int Length,int Rate);
static void DOSSetSound(int Channel,int NewType);
static void DOSDrum(int Type,int Force);
static void DOSSetChannels(int Volume,int Switch);
static void DOSSound(int Channel,int NewFreq,int NewVolume);

/** InitSound() **********************************************/
/** Initialize sound. Returns Rate on success, 0 otherwise. **/
/** Rate=0 to skip initialization (will be silent).         **/
/*************************************************************/
int InitSound(unsigned int Rate,unsigned int Latency)
{
  int I;

  /* Trash sound if it was initialized */
  TrashSound();

  /* Set driver functions */
  SndDriver.SetSound    = DOSSetSound;
  SndDriver.Drum        = DOSDrum;
  SndDriver.SetChannels = DOSSetChannels;
  SndDriver.Sound       = DOSSound;
  SndDriver.SetWave     = DOSSetWave;
  SndDriver.GetWave     = DOSGetWave;

  /* Initialize internal variables */
  SndRate      = 0;
  MasterVolume = 0;
  MasterSwitch = 0x0000;
  NoiseGen     = 1;

  /* Reset sound parameters */
  for(I=0;I<SND_CHANNELS;I++)
  {
    CH[I].Type   = SND_MELODIC;
    CH[I].Count  = 0;
    CH[I].Volume = 0;
    CH[I].Freq   = 0;
  }

  /* Rate=0 - be mute */
  if(!Rate) return(1);

  /* Done */
  SetChannels(192,(1<<SND_CHANNELS)-1);
  return(SndRate=InitAudio(Rate,Latency));
}

/** TrashSound() *********************************************/
/** Shut down sound.                                        **/
/*************************************************************/
void TrashSound(void)
{
  /* Sound trashed */
  TrashAudio();
  SndRate=0;
}

/** DOSSetSound() ********************************************/
/** Does nothing now. Supposed to set sound type (melodic/  **/
/** noise) for a channel.                                   **/
/*************************************************************/
void DOSSetSound(int Channel,int Type)
{
  /* Protect against user errors */
  if((Channel<0)||(Channel>=SND_CHANNELS)) return;
  /* Set channel type */
  CH[Channel].Type=Type;
}

/** DOSDrum() ************************************************/
/** Hit a drum of a given type with given force.            **/
/*************************************************************/
void DOSDrum(int Type,int Force)
{
  /* This function is currently empty */
}

/** DOSSound() ***********************************************/
/** Generate sound of given frequency (Hz) and volume       **/
/** (0..255) via given channel.                             **/
/*************************************************************/
void DOSSound(register int Channel,register int Freq,register int Volume)
{
  register int N;

  /* Protect against user errors */
  if((Channel<0)||(Channel>=SND_CHANNELS)) return;
  if(Freq<0) Freq=0;
  if(!Freq) Volume=0;
  Volume&=0xFF;

  CH[Channel].Volume = Volume;
  CH[Channel].Freq   = Freq;
}

/** DOSSetChannels() *****************************************/
/** Set master volume (0..255) and turn channels on/off.    **/
/** Each bit in Toggle corresponds to a channel (1=on).     **/
/*************************************************************/
void DOSSetChannels(int Volume,int Toggle)
{
  register int I;

  Volume&=0xFF;
  MasterVolume=Volume;
  MasterSwitch=Toggle;
}

/** DOSGetWave() *********************************************/
/** Get current read position for the buffer set with the   **/
/** SetWave() call. Returns 0 if no buffer has been set, or **/
/** if there is no playrate set (i.e. wave is instrument).  **/
/*************************************************************/
const signed char *DOSGetWave(int Channel)
{
  /* Channel has to be valid */
  if((Channel<0)||(Channel>=SND_CHANNELS)) return(0);
  /* Return current read position */
  return(
    CH[Channel].Rate&&(CH[Channel].Type==SND_WAVE)?
    CH[Channel].Data+CH[Channel].Pos:0
  );
}

/** DOSSetWave() *********************************************/
/** Set waveform for a given channel. The channel will be   **/
/** marked with sound type SND_WAVE. Set Rate=0 if you want **/
/** waveform to be an instrument or set it to the waveform  **/
/** own playback rate.                                      **/
/*************************************************************/
void DOSSetWave(int Channel,const signed char *Data,int Length,int Rate)
{
  /* Channel has to be valid */
  if((Channel<0)||(Channel>=SND_CHANNELS)) return;
  /* Sample length has to be valid */
  /*@@@ NOT WHEN SHARING: if((Length<=0)||(Length>SND_SAMPLESIZE)) return;*/

  /* Set channel parameters */
  CH[Channel].Type   = SND_WAVE;
  CH[Channel].Length = Length;
  CH[Channel].Rate   = Rate;
  CH[Channel].Pos    = 0;
  CH[Channel].Count  = 0;
  CH[Channel].Data   = Data;
  /*@@@ INSTEAD OF THIS: memcpy(WaveData[Channel],Data,Length);*/
}

/** RenderAudio() ********************************************/
/** Render given number of melodic sound samples. Returns   **/
/** number of samples actually rendered.                    **/
/*************************************************************/
unsigned int RenderAudio(unsigned int Samples)
{
  register int N,J,K,I,L,L1,L2,V,A1,A2;

  /* Must be using wave synthesis */
  if(SndRate<8192) return(0);

  /* Truncate to available space */
  J       = GetFreeAudio();
  Samples = Samples>SND_BUFSIZE? SND_BUFSIZE:Samples;
  Samples = Samples>J? J:Samples;
  if(!Samples) return(0);

  /* Clear mixing buffer */
  for(J=0;J<Samples;J++) Wave[J]=0;

  /* Waveform generator */
  for(J=0;J<SND_CHANNELS;J++)
    if(CH[J].Freq&&(V=CH[J].Volume)&&(MasterSwitch&(1<<J)))
      switch(CH[J].Type)
      {
        case SND_WAVE: /* Custom Waveform */
          /* Waveform data must have correct length! */
          if(CH[J].Length<=0) break;
          /* Start counting */
          K  = CH[J].Rate>0? (SndRate<<15)/CH[J].Freq/CH[J].Rate
                           : (SndRate<<15)/CH[J].Freq/CH[J].Length;
          L1 = CH[J].Pos%CH[J].Length;
          L2 = CH[J].Count;
          A1 = CH[J].Data[L1]*V;
          /* If expecting interpolation... */
          if(L2<K)
          {
            /* Compute interpolation parameters */
            A2 = CH[J].Data[(L1+1)%CH[J].Length]*V;
            L  = (L2>>15)+1;
            N  = ((K-(L2&0x7FFF))>>15)+1;
          }
          /* Add waveform to the buffer */
          for(I=0;I<Samples;I++)
            if(L2<K)
            {
              /* Interpolate linearly */
              Wave[I]+=A1+L*(A2-A1)/N;
              /* Next waveform step */
              L2+=0x8000;
              /* Next interpolation step */
              L++;
            }
            else
            {
              L1 = (L1+L2/K)%CH[J].Length;
              L2 = (L2%K)+0x8000;
              A1 = CH[J].Data[L1]*V;
              Wave[I]+=A1;
              /* If expecting interpolation... */
              if(L2<K)
              {
                /* Compute interpolation parameters */
                A2 = CH[J].Data[(L1+1)%CH[J].Length]*V;
                L  = 1;
                N  = ((K-L2)>>15)+1;
              }
            }
          /* End counting */
          CH[J].Pos   = L1;
          CH[J].Count = L2;
          break;

        case SND_NOISE: /* White Noise */
          /* For high frequencies, recompute volume */
          if(CH[J].Freq<=SndRate) K=0x10000*CH[J].Freq/SndRate;
          else { V=V*SndRate/CH[J].Freq;K=0x10000; }
          L1=CH[J].Count;
          for(I=0;I<Samples;I++)
          {
            L1+=K;
            if(L1&0xFFFF0000)
            {
              if((NoiseGen<<=1)&0x80000000) NoiseGen^=0x08000001;
              L1&=0xFFFF;
            }
            Wave[I]+=(NoiseGen&1? 127:-128)*V;
          }
          CH[J].Count=L1;
          break;

        case SND_MELODIC:  /* Melodic Sound   */
        case SND_TRIANGLE: /* Triangular Wave */
        default:           /* Default Sound   */
          /* Do not allow frequencies that are too high */
          if(CH[J].Freq>=SndRate/2) break;
          K=0x10000*CH[J].Freq/SndRate;
          L1=CH[J].Count;
          for(I=0;I<Samples;I++,L1+=K)
            Wave[I]+=(L1&0x8000? 127:-128)*V;
          CH[J].Count=L1&0xFFFF;
          break;
      }

  /* Mix and convert waveforms */
  for(J=0;J<Samples;J++)
  {
    I         = (Wave[J]*MasterVolume)/255;
    I         = I>32767? 32767:I<-32768? -32768:I;
#ifdef BPS16
    OutBuf[J] = I;
#else
    OutBuf[J] = I>>8;
#endif
  }

  /* Write audio */
  return(WriteAudio(OutBuf,Samples));
}
