/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                          SndWin.c                       **/
/**                                                         **/
/** This file contains standard sound generation routines   **/
/** for Windows using wave synthesis or MIDI.               **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-2007                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#include "EMULib.h"
#include "Sound.h"

/** SHIFT_CHANNEL() ******************************************/
/** Make MIDI channel#10 last, as it is normally used for   **/
/** percussion instruments only and doesn't sound nice.     **/
/*************************************************************/
#define SHIFT_CHANNEL(Ch) (Ch==15? 9:Ch>8? Ch+1:Ch)

static const struct { BYTE Note;WORD Wheel; } Freqs[4096] =
{
#include "MIDIFreq.h"
};

int MasterVolume = 200;                  /* These are made public so  */
int MasterSwitch = (1<<SND_CHANNELS)-1;  /* that LibWin.c can access  */
static int SndRate      = 0;
static int NoiseGen;
static int DrumOn;
static HMIDIOUT hMIDI   = 0;
static int Wave[SND_BUFSIZE];
static sample OutBuf[SND_BUFSIZE];

static struct
{
  int Type;                       /* Channel type (SND_*)             */
  int Freq;                       /* Channel frequency (Hz)           */
  int Volume;                     /* Channel volume (0..255)          */

  int Note;                       /* MIDI note (-1 for off)           */
  int Pitch;                      /* MIDI pitch                       */
  int Level;                      /* MIDI volume level (velocity)     */

  const signed char *Data;        /* Wave data (-128..127 each)       */
  int Length;                     /* Wave length in Data              */
  int Rate;                       /* Wave playback rate (or 0Hz)      */
  int Pos;                        /* Wave current position in Data    */

  int Count;                      /* Phase counter                    */
} CH[SND_CHANNELS] =
{
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 },
  { SND_MELODIC,0,0,-1,-1,-1,0,0,0,0 }
};

static void WinSetSound(int Channel,int Type);
static void WinDrum(int Type,int Force);
static void WinSetChannels(int Volume,int Switch);
static const signed char *WinGetWave(int Channel);
static void WinSetWave(int Channel,const signed char *Data,int Length,int Rate);
static void WinSound(int Channel,int Freq,int Volume);
static int WinInitWave(HINSTANCE hInst,int Rate,int Buffers);
static int WinInitMIDI(HINSTANCE hInst);
static void NoteOn(int Channel,int Note,int Level);
static void NoteOff(int Channel);

/** InitSound() **********************************************/
/** Initialize sound. Returns 0 on failure, effective rate  **/
/** otherwise. Special cases of rate argument: 0 - return 1 **/
/** and be silent, 1..8191 - use MIDI mapper to play music. **/
/*************************************************************/
unsigned int InitSound(unsigned int Rate,unsigned int Delay)
{
  int J;

  /* If initialized, shut down */
  TrashSound();

  /* Set driver functions */
  SndDriver.SetSound    = WinSetSound;
  SndDriver.Drum        = WinDrum;
  SndDriver.SetChannels = WinSetChannels;
  SndDriver.Sound       = WinSound;
  SndDriver.SetWave     = WinSetWave;
  SndDriver.GetWave     = WinGetWave;

  /* Reset channels */
  for(J=0;J<SND_CHANNELS;J++)
  {
    CH[J].Freq   = 0;
    CH[J].Volume = 0;
    CH[J].Count  = 0;
    CH[J].Note   = -1;
    CH[J].Pitch  = -1;
    CH[J].Level  = -1;
  }

  /* Reset variables */
  hMIDI    = 0;
  NoiseGen = 1;
  SndRate  = 0;
  DrumOn   = 0;

  /* No sound requested: EXIT */
  if(!Rate) return(1);

  /* Reopen sound */
  SndRate=Rate<8192? WinInitMIDI(WinGetAppInstance()):InitAudio(Rate,Delay);

  /* Fall out if failed */
  if(!SndRate) return(0);

  /* Set current instrments (if reopening device) */
  for(J=0;J<SND_CHANNELS;J++) WinSetSound(J,CH[J].Type);

  /* Done */
  return(SndRate);
}

/** TrashSound() *********************************************/
/** Close all devices and free all allocated resources.     **/
/*************************************************************/
void TrashSound(void)
{
  if(SndRate)
  {
    /* Shutdown wave audio or close midiOut */
    if(SndRate>=8192) TrashAudio();
    else
      if(hMIDI)
      {
        midiOutReset(hMIDI);
        midiOutClose(hMIDI);
        hMIDI=0;
      }

    /* Done */
    SndRate=0;
  }
}

/** WinSetSound() ********************************************/
/** Set sound type for a given channel.                     **/
/*************************************************************/
void WinSetSound(int Channel,int Type)
{
  static int Programs[5] = { 80,80,122,122,80 };

  if(!SndRate||(Channel<0)||(Channel>=SND_CHANNELS)) return;

  CH[Channel].Type=Type;

  if(SndRate<8192)
  {
    Type=Type&SND_MIDI? Type&0x7F:Programs[Type&0x03];
    midiOutShortMsg(hMIDI,(Type<<8)|(SHIFT_CHANNEL(Channel)&0x0F)|0xC0);
  }
}

/** WinDrum() ************************************************/
/** Hit a drum of a given type with given force.            **/
/*************************************************************/
void WinDrum(int Type,int Force)
{
  if(!SndRate) return;
  Force=Force<0? 0:Force>255? 255:Force;

  if((SndRate<8192)&&(MasterSwitch&0x8000))
  {
    /* The only non-MIDI drum is a click ("Low Wood Block") */
    Type=Type&DRM_MIDI? Type&0x7F:77;
    /* Release current drum */
    if(DrumOn) midiOutShortMsg(hMIDI,(127<<16)|(Type<<8)|0x89);
    /* Hit current drum */
    if(Type) midiOutShortMsg(hMIDI,(((Force+1)>>1)<<16)|(Type<<8)|0x99);
    DrumOn=Type;
  }
}

/** WinSound() ***********************************************/
/** Set sound volume and frequency for a given channel.     **/
/*************************************************************/
void WinSound(int Channel,int Freq,int Volume)
{
  int MIDIVolume,MIDINote,MIDIWheel;

  if(!SndRate||(Channel<0)||(Channel>=SND_CHANNELS)) return;
  Volume=Volume<0? 0:Volume>255? 255:Volume;

  if(SndRate>=8192)
  {
    /* Store frequency and volume */
    CH[Channel].Freq   = Freq;
    CH[Channel].Volume = Volume;
  }
  else
  {
    /* Clip the frequency */
    if((Freq<MIDI_MINFREQ)||(Freq>MIDI_MAXFREQ)) Freq=0;
    /* SND_TRIANGLE has 1/2 volume of SND_MELODIC */
    if(CH[Channel].Type==SND_TRIANGLE) Volume=(Volume+1)/2;
    /* Store frequency and volume */
    CH[Channel].Freq   = Freq;
    CH[Channel].Volume = Volume;

    if(MasterSwitch&(1<<Channel))
    {
      if(!Volume||!Freq) NoteOff(Channel);
      else
      {
        /* MIDI volume is 0..127, frequency is 0..4095 */
        MIDIVolume = (127*Volume+128)/255;
        MIDINote   = Freqs[Freq/3].Note;
        MIDIWheel  = Freqs[Freq/3].Wheel;

        /* Play new note */
        NoteOn(Channel,MIDINote,MIDIVolume);

        /* Change pitch */
        if(CH[Channel].Pitch!=MIDIWheel)
        {
          midiOutShortMsg
          (
            hMIDI,
            0xE0
            | (SHIFT_CHANNEL(Channel)&0x0F)
            | ((MIDIWheel&0x3F80)<<1)
            | ((MIDIWheel&0x7F)<<8)
          );
          CH[Channel].Pitch=MIDIWheel;
        }
      }
    }
  }
}

/** WinSetChannels() *****************************************/
/** Set overall sound volume and switch channels on/off.    **/
/*************************************************************/
void WinSetChannels(int Volume,int Switch)
{
  int J;

  if(!SndRate) return;
  Volume=Volume<0? 0:Volume>255? 255:Volume;
  Switch&=(1<<SND_CHANNELS)-1;

  if(SndRate<8192)
  {
    for(J=0;J<SND_CHANNELS;J++)
      if(MasterSwitch&(Switch^MasterSwitch)&(1<<J)) NoteOff(J);
    /*@@@ NO HW VOLUME midiOutSetVolume((UINT)hMIDI,Volume*0x1000100L);*/
  }

  MasterVolume=Volume;
  MasterSwitch=Switch;
}

/** GetWave() ************************************************/
/** Get current read position for the buffer set with the   **/
/** SetWave() call. Returns 0 if no buffer has been set, or **/
/** if there is no playrate set (i.e. wave is instrument).  **/
/*************************************************************/
const signed char *WinGetWave(int Channel)
{
  /* Channel has to be valid */
  if((Channel<0)||(Channel>=SND_CHANNELS)) return(0);
  /* Return current read position */
  return(
    CH[Channel].Rate&&(CH[Channel].Type==SND_WAVE)?
    CH[Channel].Data+CH[Channel].Pos:0
  );
}

/** SetWave() ************************************************/
/** Set waveform for a given channel. The channel will be   **/
/** marked with sound type SND_WAVE. Set Rate=0 if you want **/
/** waveform to be an instrument or set it to the waveform  **/
/** own playback rate.                                      **/
/*************************************************************/
void WinSetWave(int Channel,const signed char *Data,int Length,int Rate)
{
  /* SetWave() call only makes sense for waveOut sound, not midiOut */
  if(SndRate<8192) { WinSetSound(Channel,SND_MELODIC);return; }
  /* Channel has to be valid */
  if((Channel<0)||(Channel>=SND_CHANNELS)) return;

  /* Set channel parameters */
  CH[Channel].Type   = SND_WAVE;
  CH[Channel].Length = Length;
  CH[Channel].Rate   = Rate;
  CH[Channel].Data   = Data;
  CH[Channel].Pos    = 0;
  CH[Channel].Count  = 0;
}

/** NoteOn() *************************************************/
/** Play a new note.                                        **/
/*************************************************************/
void NoteOn(int Channel,int Note,int Level)
{
  Note  = Note<0? 0x00:Note>0x7F? 0x7F:Note;
  Level = Level<0? 0x00:Level>0x7F? 0x7F:Level;

  if(hMIDI&&((Note!=CH[Channel].Note)||(Level!=CH[Channel].Level)))
  {
    if(CH[Channel].Note>=0) NoteOff(Channel);
    midiOutShortMsg
    (
      hMIDI,
      0x90
      | (SHIFT_CHANNEL(Channel)&0x0F)
      | (Level<<16)
      | (Note<<8)
    );
    CH[Channel].Note  = Note;
    CH[Channel].Level = Level;
  }
}

/** NoteOff() ************************************************/
/** Turn off a note.                                        **/
/*************************************************************/
void NoteOff(int Channel)
{
  if(hMIDI&&(CH[Channel].Note>=0))
  {
    midiOutShortMsg
    (
      hMIDI,
      0x80
      | (SHIFT_CHANNEL(Channel)&0x0F)
      | ((CH[Channel].Note&0x7F)<<8)
    );
    CH[Channel].Note=-1;
  }
}

/** WinInitMIDI() ********************************************/
/** Initialize midiOut. Returns 0 on failure or 1 otherwise **/
/*************************************************************/
int WinInitMIDI(HINSTANCE hInst)
{
  DWORD Data;
  int J;

  /* No sound yet */
  SndRate=0;

  /* Open MIDI mapper */
  if(midiOutOpen(&hMIDI,MIDI_MAPPER,0,0,0)) return(0);

  /* midiOut opened successfully */
  SndRate=1;

  /* Set initial volumes */
 // for(J=0;J<SND_CHANNELS;J++) midiOutShortMsg(hMIDI,0x6407B0|J);

  /* Success! */
  return(SndRate);
}

/** RenderAudio() ********************************************/
/** Render given number of melodic sound samples. Returns   **/
/** number of samples actually rendered.                    **/
/*************************************************************/
unsigned int RenderAudio(unsigned int Samples)
{
  register int N,J,K,I,L,L1,L2,V,A1,A2;

  /* Must be using wave synthesis */
  if(SndRate<8192) return(0);

  /* Truncate to available space */
  J       = GetFreeAudio();
  Samples = Samples>SND_BUFSIZE? SND_BUFSIZE:Samples;
  Samples = Samples>J? J:Samples;
  if(!Samples) return(0);

  /* Clear mixing buffer */
  for(J=0;J<Samples;J++) Wave[J]=0;

  /* Waveform generator */
  for(J=0;J<SND_CHANNELS;J++)
    if(CH[J].Freq&&(V=CH[J].Volume)&&(MasterSwitch&(1<<J)))
      switch(CH[J].Type)
      {
        case SND_WAVE: /* Custom Waveform */
          /* Waveform data must have correct length! */
          if(CH[J].Length<=0) break;
          /* Start counting */
          K  = CH[J].Rate>0? (SndRate<<15)/CH[J].Freq/CH[J].Rate
                           : (SndRate<<15)/CH[J].Freq/CH[J].Length;
          L1 = CH[J].Pos%CH[J].Length;
          L2 = CH[J].Count;
          A1 = CH[J].Data[L1]*V;
          /* If expecting interpolation... */
          if(L2<K)
          {
            /* Compute interpolation parameters */
            A2 = CH[J].Data[(L1+1)%CH[J].Length]*V;
            L  = (L2>>15)+1;
            N  = ((K-(L2&0x7FFF))>>15)+1;
          }
          /* Add waveform to the buffer */
          for(I=0;I<Samples;I++)
            if(L2<K)
            {
              /* Interpolate linearly */
              Wave[I]+=A1+L*(A2-A1)/N;
              /* Next waveform step */
              L2+=0x8000;
              /* Next interpolation step */
              L++;
            }
            else
            {
              L1 = (L1+L2/K)%CH[J].Length;
              L2 = (L2%K)+0x8000;
              A1 = CH[J].Data[L1]*V;
              Wave[I]+=A1;
              /* If expecting interpolation... */
              if(L2<K)
              {
                /* Compute interpolation parameters */
                A2 = CH[J].Data[(L1+1)%CH[J].Length]*V;
                L  = 1;
                N  = ((K-L2)>>15)+1;
              }
            }
          /* End counting */
          CH[J].Pos   = L1;
          CH[J].Count = L2;
          break;

        case SND_NOISE: /* White Noise */
          /* For high frequencies, recompute volume */
          if(CH[J].Freq<=SndRate) K=0x10000*CH[J].Freq/SndRate;
          else { V=V*SndRate/CH[J].Freq;K=0x10000; }
          L1=CH[J].Count;
          for(I=0;I<Samples;I++)
          {
            L1+=K;
            if(L1&0xFFFF0000)
            {
              if((NoiseGen<<=1)&0x80000000) NoiseGen^=0x08000001;
              L1&=0xFFFF;
            }
            Wave[I]+=(NoiseGen&1? 127:-128)*V;
          }
          CH[J].Count=L1;
          break;

        case SND_MELODIC:  /* Melodic Sound   */
        case SND_TRIANGLE: /* Triangular Wave */
        default:           /* Default Sound   */
          /* Do not allow frequencies that are too high */
          if(CH[J].Freq>=SndRate/2) break;
          K=0x10000*CH[J].Freq/SndRate;
          L1=CH[J].Count;
          for(I=0;I<Samples;I++,L1+=K)
            Wave[I]+=(L1&0x8000? 127:-128)*V;
          CH[J].Count=L1&0xFFFF;
          break;
      }

  /* Mix and convert waveforms */
  for(J=0;J<Samples;J++)
  {
    I         = (Wave[J]*MasterVolume)/255;
    I         = I>32767? 32767:I<-32768? -32768:I;
#ifdef BPS16
    OutBuf[J] = I;
#else
    OutBuf[J] = I>>8;
#endif
  }

  /* Write audio */
  return(WriteAudio(OutBuf,Samples));
}
